/*
 * Copyright (c) 2001 Massachusetts Institute of Technology.
 *
 * This file is part of VideoCM. 
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `COPYING' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file AUTHORS which should have accompanied this software 
 * for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the AUTHORS file.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Network Research
 *	Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */


#include "pktbuf.h"
#include "rtp.h"
#include "pktbuf-rtp.h"

RTP_BufferPool_::RTP_BufferPool_() : freebufs_(0), nbufs_(0), nused_(0)
{
}

RTP_pktbuf* RTP_BufferPool_::alloc(unsigned char type)
{
  RTP_pktbuf* pb = freebufs_;
  if (pb != 0) {
#ifdef SRRTP_DEBUG_ENABLED
    fprintf(stderr, "Free Buffers: %d\n", nbufs_ - nused_);
#endif
    freebufs_ = pb->next;
  } else {
    /*XXX grow exponentially*/
#ifdef SRRTP_DEBUG_ENABLED
    fprintf(stderr, "ALLOCATING (%d, %d)\n", nbufs_, nbufs_-nused_);
#endif
    pb = new RTP_pktbuf;
    pb->manager = this;
    ++nbufs_;
  }
  pb->len = 0;
  pb->offset = 0; 
  pb->dp = pb->data;
  nused_++;
  return (pb);
}

void RTP_BufferPool_::release(RTP_pktbuf* pb)
{
  pb->next  = freebufs_;
  freebufs_ = pb;
  nused_--;
#ifdef SRRTP_DEBUG_ENABLED
  fprintf(stderr, "FREEING(%d, %d)\n", nbufs_, nbufs_-nused_);
#endif
}

RTP_pktbuf* RTP_pktbuf::copy(u_char type)
{
  RTP_pktbuf* cp;
  cp = ((RTP_BufferPool_ *)manager)->alloc(type);
  memcpy(cp->dp, dp, len);
  cp->len += len;
  cp->offset = offset;
  return (cp);
}


void
RTP_pktbuf::update_rrtp_timestamp() {
  rrtphdr *rh = (rrtphdr*)data;
  rh->rh_ts = htonl(ntptime());
}
